import os
from functools import partial

from generate_cmake.log import Log
from generate_cmake.file_system import reduce_srcs


def should_create_target(build_item):
    if 'NOPYMAKE' in os.listdir(build_item['location']):
        Log.debug("Ignoring: {}".format(build_item['target']))
        return False
    else:
        return True


def create_lib(build_item, base_directory=""):
    bloated_srcs = map(partial(os.path.join, build_item['location']), build_item['srcs'])
    # txt = "add_library({} {})".format(build_item['target'], " ".join(reduced_srcs))

    if(len(build_item['deps'])):
        import IPython; IPython.embed(); exit(0)

    txt = """
cc_library(
name = {name},
srcs = [{sources}],
hdrs = [{headers}],
deps = [{deps}]
visibility = ["//visibility:public"]
)
""".format(
        name=build_item['target'],
        sources=", ".join(build_item['hdrs']),
        headers=", ".join(build_item['hdrs'])
        # deps=""
    )

    # If the library has dependencies
    if len(build_item['deps']):
        txt += "\ntarget_link_libraries({} {})".format(build_item['target'], " ".join(build_item['deps']))
    return txt


def create_bin(build_item, base_directory=""):
    reduced_srcs = reduce_srcs(build_item['srcs'], base_directory)

    if 'is_test' in build_item.get('flags', []):
        assert len(reduced_srcs) == 1, "A test only supports one source file"
        txt = 'add_test({target} {srcs} {libs})'.format(
            target=build_item['target'],
            srcs=reduced_srcs[0],
            libs=" ".join(build_item['deps']) if len(build_item['deps']) else '""'
        )
    else:
        txt = "add_executable({} {})".format(build_item['target'], " ".join(reduced_srcs))
        if len(build_item['deps']):
            txt += "\ntarget_link_libraries({} {})".format(build_item['target'], " ".join(build_item['deps']))
    return txt


def build_bazel_build(to_build, base_directory):
    actions = {
        'lib': create_lib,
        'binary': create_bin
    }

    from graph import dependency_sort
    write_order = dependency_sort(to_build)

    Log.success('\n\nGenerating BUILDs....')
    cmake_text = ""
    for write in write_order:
        if write in to_build:
            build_item = to_build[write]
            if not should_create_target(build_item):
                continue

            action = actions[build_item['kind']]
            new_text = action(build_item, base_directory=base_directory)
            cmake_text += new_text + '\n'
            Log.info(new_text)

    # write_build(cmake_text, base_directory)



import os
import re

from generate_cmake.parse import between

valid_extensions = ['cc', 'hh']


def parse_cfg(path, base):
    with open(os.path.join(path, 'cfg.pymake')) as f:
        text = f.read()

    include_paths = []
    commands = text.split('\n')
    for command in commands:
        if(command.startswith('add_path') or
           command.startswith('add_ext_path')):
            include_paths.append(
                os.path.join(
                    base,
                    between(command, '(', ')')
                )
            )

    return {'include_paths': include_paths}


def resolve_include_TODO(file_path, include_path, available_include_paths):
    dirname = os.path.dirname(file_path)
    other_files = filter(os.path.isfile, os.listdir(dirname))

    if include_path in other_files:
        return os.path.join(dirname, include_path)
    # TODO: Make this handle more cases


def resolve_include(file_path, include_path, available_include_paths=[]):
    # return os.path.join('/home/jacob/repos/experiments', include_path)
    directory, _ = os.path.split(file_path)
    extended_available_incl_paths = available_include_paths + [directory]
    for path in extended_available_incl_paths:
        potential_path = os.path.join(path, include_path)
        if os.path.exists(potential_path):
            return potential_path
    else:
        raise ValueError("Unknown include: {} in {}".format(include_path, file_path))


def parse_ignore(path):
    if path is None or '.ignore' not in os.listdir(path):
        ignores = [
            ".git",
            "bin/",
            "third_party/",
            "common/",
        ]
    else:
        with open(os.path.join(path, '.ignore')) as f:
            text = f.read()
        ignores = text.split('\n')
    return list(map(re.compile, ignores))


def deep_listdir(path):
    dir_contents = os.listdir(path)
    return map(lambda o: os.path.join(path, o), dir_contents)


def get_files(path, ignores_path=None, ignores=None):
    from functools import partial
    if ignores is None:
        ignores = parse_ignore(ignores_path)

    def want_file(file):
        re_partial = partial(re.match, string=file)
        allow_by_is_file = os.path.isfile(file)
        allow_by_extension = any(map(file.endswith, valid_extensions))
        allow_by_ignores = not any(map(re_partial, ignores))

        directory, _ = os.path.split(file)
        requires = [
            allow_by_ignores,
            allow_by_extension,
            allow_by_is_file,
        ]

        return all(requires)

    def want_dir(dirname):
        directory, filename = os.path.split(dirname)
        re_partial = partial(re.match, string=filename + '/')
        allow_by_ignores = not any(map(re_partial, ignores))
        return allow_by_ignores and os.path.isdir(dirname)

    dir_elements = deep_listdir(path)
    files_here = filter(want_file, dir_elements)
    dirs_here = filter(want_dir, dir_elements)

    for subdir in dirs_here:
        subdir_path = os.path.join(path, subdir)
        if 'CMakeLists.txt' in os.listdir(subdir_path):
            continue

        files_there = get_files(subdir_path, ignores=ignores)
        files_here.extend(files_there)
    return files_here


def reduce_srcs(sources, base_directory, prejoin="${PROJECT_SOURCE_DIR}"):
    # return map(lambda o: os.path.join(prejoin, o.replace(base_directory + '/', "")), sources)
    # return map(lambda o: os.path.join(prejoin, o.replace(base_directory + '/', "")), sources)
    return sources



from generate_cmake.log import Log
from functools import partial


def splitstrip(text, delim=','):
    return list(map(lambda o: o.strip(), text.split(delim)))


def needs(text, requirements):
    for requirement in requirements:
        assert requirement in text, "{} missing {}".format(text, requirement)


def mustnot(text, disquirements):
    for requirement in disquirements:
        assert requirement not in text, "{} may not have {}".format(text, requirement)


def between(text, ldelim='(', rdelim=')'):
    needs(text, [ldelim, rdelim])
    start = text.find(ldelim)
    end = text.rfind(rdelim)
    assert end - start > 1, "Delimeters empty in: {}".format(text)
    return text[start + 1: end]


pairs = {
    '<': '>',
    '(': ')',
    '"': '"',
    "'": "'",
}


def grab_dependencies(line):
    '''Parse dependencies out of a line'''
    needs(line, ['(', ')'])
    needs(line, ['//%deps'])

    args = between(line, '(', ')')
    deps = splitstrip(args)
    Log.debug('    deps: {}'.format(deps))
    return {'deps': deps}


def grab_lib(line):
    needs(line, ['(', ')'])
    needs(line, ['//%lib'])
    args = between(line, '(', ')')
    arg_tokens = splitstrip(args)
    Log.debug('    lib: {}'.format(arg_tokens))
    return {'lib': {'target': arg_tokens[0], 'srcs': arg_tokens[1:]}}


def grab_bin(line):
    needs(line, ['(', ')'])
    needs(line, ['//%bin'])
    args = between(line, '(', ')')
    arg_tokens = splitstrip(args)
    Log.debug('    bin: {}'.format(arg_tokens))
    assert len(arg_tokens) in (0, 1), "Wrong number of arguments"
    return {'bin': {'target': arg_tokens[0]}}


def grab_include(line, incl_type='"'):
    needs(line, ['#include'])

    incl_path = ""
    for start, end in pairs.items():
        if start in line:
            incl_path = between(line, start, end)
            break

    if (incl_type not in ("local", "system")):
        raise NotImplementedError("wtf?")

    Log.debug("    include: {} ({})".format(incl_path, incl_type))

    return {'include': {'given_path': incl_path, 'type': incl_type}}


def make_flagger(flag_name):
    def flagger(line):
        Log.debug('    flag: {}'.format(flag_name))
        return {'flags': [flag_name]}
    return flagger


tokens = {
    '//%bin': grab_bin,
    '//%deps': grab_dependencies,
    '//%lib': grab_lib,
    '#include <': partial(grab_include, incl_type="system"),
    '#include "': partial(grab_include, incl_type="local"),
    'int main(': make_flagger('has_main'),
    'void main(': make_flagger('has_main'),
    '//%ignore': make_flagger('ignore'),
    '#include "testing/gtest.hh"': make_flagger('is_test'),
    '//%test': make_flagger('is_test'),
}


def incremental_update(d, new_d):
    for key, item in new_d.items():
        if key not in d:
            d[key] = []

        if isinstance(item, list):
            d[key].extend(item)
        else:
            d[key].append(item)


def strip_spaces_after_comment(dirty_line):
    if dirty_line.startswith('//'):
        return '//' + dirty_line[2:].lstrip()
    else:
        return dirty_line


def parse_text(text):
    elements = {
        'bin': [],
        'deps': [],
        'lib': [],
        'include': [],
        'flags': []
    }

    lines = text.split('\n')
    for dirty_line in lines:
        line = strip_spaces_after_comment(dirty_line)
        for token, action in tokens.items():
            if line.startswith(token):
                update = action(line)
                incremental_update(elements, update)

    return elements


def has_annotations(elements):
    no_deps = len(elements['deps']) == 0
    no_bin = len(elements['bin']) == 0
    no_lib = len(elements['lib']) == 0
    return not all([no_deps, no_bin, no_lib])


def should_ignore(elements):
    return 'ignore' in elements['flags']


def main():
    text = "// %deps(a, b, c)\n//    %deps(ogop)\n//%deps(abs)\n%deps(adq)"
    parsed = parse_text(text)
    assert set(parsed['deps']) == {'a', 'b', 'c', 'ogop', 'abs'}


if __name__ == '__main__':
    main()
